@@ -52,8 +52,6 @@
 	 pkix_path_validation/3,
 	 pkix_verify_hostname/2, pkix_verify_hostname/3,
          pkix_verify_hostname_match_fun/1,
-	 ssh_decode/2, ssh_encode/2,
-	 ssh_hostkey_fingerprint/1, ssh_hostkey_fingerprint/2,
 	 ssh_curvename2oid/1, oid2ssh_curvename/1,
 	 pkix_crls_validate/3,
 	 pkix_dist_point/1,
@@ -69,6 +67,34 @@
          ocsp_extensions/1
 	]).
 
+%%----------------
+%% To be moved to ssh and deprecated:
+-export([ssh_decode/2, ssh_encode/2,
+         ssh_hostkey_fingerprint/1, ssh_hostkey_fingerprint/2
+        ]).
+
+-deprecated([{ssh_decode,2, "use ssh_file:decode/2 instead"},
+             {ssh_encode,2, "use ssh_file:encode/2 instead"},
+             {ssh_hostkey_fingerprint,1, "use ssh:hostkey_fingerprint/1 instead"},
+             {ssh_hostkey_fingerprint,2, "use ssh:hostkey_fingerprint/2 instead"}
+            ]).
+
+-compile([{nowarn_deprecated_function,
+           [{public_key,ssh_decode,2},
+            {public_key,ssh_encode,2}
+           ]}
+         ]).
+
+%% When removing for OTP-25.0, remember to also remove
+%%   - most of pubkey_ssh.erl except
+%%       + dh_gex_group/4
+%%       + dh_gex_group_sizes/0
+%%   - pubkey_pem:pem_start({no_asn1, new_openssh})
+%%   - pubkey_pem:pem_end(<<"-----BEGIN OPENSSH PRIVATE KEY-----">>)
+%%   - pubkey_pem:asn1_type(<<"-----BEGIN OPENSSH PRIVATE KEY-----">>) -> {no_asn1, new_openssh}
+
+%%----------------------------------------------------------------
+%% Types
 -export_type([public_key/0, private_key/0, pem_entry/0,
 	      pki_asn1_type/0, asn1_type/0, ssh_file/0, der_encoded/0,
               key_params/0, digest_type/0, issuer_name/0, cert_id/0, oid/0]).
@@ -85,9 +111,10 @@
 -type ecpk_parameters_api() :: ecpk_parameters() | #'ECParameters'{} | {namedCurve, Name::crypto:ec_named_curve()}.
 -type ec_public_key()        :: {#'ECPoint'{}, ecpk_parameters_api()}.
 -type ec_private_key()       :: #'ECPrivateKey'{}.
--type ed_public_key()        :: {ed_pub, ed25519|ed448, Key::binary()}.
--type ed_private_key()       :: {ed_pri, ed25519|ed448, Pub::binary(), Priv::binary()}.
-
+-type ed_public_key()        :: {#'ECPoint'{}, ed_params()} | {ed_pub, ed25519|ed448, Key::binary()}.
+-type ed_private_key()       :: #'ECPrivateKey'{parameters :: ed_params()} |
+                                {ed_pri, ed25519|ed448, Pub::binary(), Priv::binary()}.
+-type ed_params()            ::  {namedCurve, ?'id-Ed25519' | ?'id-Ed448'}.
 -type key_params()           :: #'DHParameter'{} | {namedCurve, oid()} | #'ECParameters'{} | 
                                 {rsa, Size::integer(), PubExp::integer()}. 
 -type der_encoded()          :: binary().
@@ -109,11 +136,6 @@
                               | #'PBES2-params'{} .
 
 -type salt()                 :: binary(). % crypto:strong_rand_bytes(8)
-%% -type cipher_info()          :: {Cipher :: string(), Salt :: binary()} |
-%%                                 {Cipher :: string(), #'PBES2-params'{}} | 
-%%                                 {Cipher :: string(), {#'PBEParameter'{}, atom()}} %% hash type
-%%                                 .
-
 -type asn1_type()            :: atom(). %% see "OTP-PUB-KEY.hrl
 -type ssh_file()             :: openssh_public_key | rfc4716_public_key | known_hosts |
 				auth_keys.
@@ -128,7 +150,7 @@
 -type chain_type()           :: server_chain | client_chain.
 
 -type cert_id()              :: {SerialNr::integer(), issuer_name()} .
--type issuer_name()          :: {rdnSequence,[#'AttributeTypeAndValue'{}]} .
+-type issuer_name()          :: {rdnSequence,[[#'AttributeTypeAndValue'{}]]} .
 
 
 
@@ -215,23 +237,32 @@ pem_entry_decode({Asn1Type, CryptDer, {Cipher, Salt}} = PemEntry,
                                                              Entity :: term() .
 
 pem_entry_encode('SubjectPublicKeyInfo', Entity=#'RSAPublicKey'{}) ->
-    Der = der_encode('RSAPublicKey', Entity),
-    Spki = {'SubjectPublicKeyInfo',
-            {'AlgorithmIdentifier', ?'rsaEncryption', ?DER_NULL}, Der},
+    KeyDer = der_encode('RSAPublicKey', Entity),
+    Spki = subject_public_key_info(#'AlgorithmIdentifier'{algorithm = ?'rsaEncryption',
+                                                          parameters =?DER_NULL}, KeyDer),
     pem_entry_encode('SubjectPublicKeyInfo', Spki);
 pem_entry_encode('SubjectPublicKeyInfo',
                  {DsaInt, Params=#'Dss-Parms'{}}) when is_integer(DsaInt) ->
     KeyDer = der_encode('DSAPublicKey', DsaInt),
     ParamDer = der_encode('DSAParams', {params, Params}),
-    Spki = {'SubjectPublicKeyInfo',
-            {'AlgorithmIdentifier', ?'id-dsa', ParamDer}, KeyDer},
+    Spki = subject_public_key_info(#'AlgorithmIdentifier'{algorithm =?'id-dsa',
+                                                          parameters = ParamDer},
+                                   KeyDer),
+    pem_entry_encode('SubjectPublicKeyInfo', Spki);
+pem_entry_encode('SubjectPublicKeyInfo',
+		 {#'ECPoint'{point = Key}, {namedCurve, ?'id-Ed25519' = ID}}) when is_binary(Key)->
+    Spki = subject_public_key_info(#'AlgorithmIdentifier'{algorithm = ID}, Key),
+    pem_entry_encode('SubjectPublicKeyInfo', Spki);
+pem_entry_encode('SubjectPublicKeyInfo',
+		 {#'ECPoint'{point = Key}, {namedCurve, ?'id-Ed448' = ID}}) when is_binary(Key)->
+    Spki = subject_public_key_info(#'AlgorithmIdentifier'{algorithm = ID}, Key),
     pem_entry_encode('SubjectPublicKeyInfo', Spki);
 pem_entry_encode('SubjectPublicKeyInfo',
 		 {#'ECPoint'{point = Key}, ECParam}) when is_binary(Key)->
     Params = der_encode('EcpkParameters',ECParam),
-    Spki = {'SubjectPublicKeyInfo',
-	    {'AlgorithmIdentifier', ?'id-ecPublicKey', Params},
-	    Key},
+    Spki = subject_public_key_info(#'AlgorithmIdentifier'{algorithm =?'id-ecPublicKey',
+                                                          parameters = Params},
+                                   Key),
     pem_entry_encode('SubjectPublicKeyInfo', Spki);
 pem_entry_encode(Asn1Type, Entity)  when is_atom(Asn1Type) ->
     Der = der_encode(Asn1Type, Entity),
@@ -270,15 +301,16 @@ pem_entry_encode(Asn1Type, Entity, {{Cipher, Salt} = CipherInfo,
 %%
 %% Description: Decodes a public key asn1 der encoded entity.
 %%--------------------------------------------------------------------
-der_decode(Asn1Type, Der) when (Asn1Type == 'PrivateKeyInfo') or 
-			       (Asn1Type == 'EncryptedPrivateKeyInfo')
-			       andalso is_binary(Der) ->
+der_decode(Asn1Type, Der) when (((Asn1Type == 'PrivateKeyInfo')
+                                 orelse
+                                   (Asn1Type == 'EncryptedPrivateKeyInfo'))
+                                andalso is_binary(Der)) ->
     try
 	{ok, Decoded} = 'PKCS-FRAME':decode(Asn1Type, Der),
 	der_priv_key_decode(Decoded)
     catch
 	error:{badmatch, {error, _}} = Error ->
-	    erlang:error(Error)
+            handle_pkcs_frame_error(Asn1Type, Der, Error)
     end;
 
 der_decode(Asn1Type, Der) when is_atom(Asn1Type), is_binary(Der) ->
@@ -290,10 +322,40 @@ der_decode(Asn1Type, Der) when is_atom(Asn1Type), is_binary(Der) ->
 	    erlang:error(Error)
     end.
 
-der_priv_key_decode({'PrivateKeyInfo', v1,
-	{'PrivateKeyInfo_privateKeyAlgorithm', ?'id-ecPublicKey', {asn1_OPENTYPE, Parameters}}, PrivKey, _}) ->
-	EcPrivKey = der_decode('ECPrivateKey', PrivKey),
-	EcPrivKey#'ECPrivateKey'{parameters = der_decode('EcpkParameters', Parameters)};
+handle_pkcs_frame_error('PrivateKeyInfo', Der, _) ->
+    try
+	{ok, Decoded} = 'PKCS-FRAME':decode('OneAsymmetricKey', Der),
+	der_priv_key_decode(Decoded)
+    catch
+	error:{badmatch, {error, _}} = Error ->
+	    erlang:error(Error)
+    end;
+handle_pkcs_frame_error(_, _, Error) ->
+    erlang:error(Error).
+
+der_priv_key_decode(#'PrivateKeyInfo'{version = v1,
+                                      privateKeyAlgorithm =
+                                          #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = ?'id-ecPublicKey',
+                                                                                parameters = {asn1_OPENTYPE, Parameters}},
+                                      privateKey = PrivKey}) ->
+    EcPrivKey = der_decode('ECPrivateKey', PrivKey),
+    EcPrivKey#'ECPrivateKey'{parameters = der_decode('EcpkParameters', Parameters)};
+der_priv_key_decode(#'PrivateKeyInfo'{version = v1,
+                                      privateKeyAlgorithm =#'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = CurveOId},
+                                      privateKey = PrivKey}) when
+      CurveOId == ?'id-Ed25519'orelse
+      CurveOId == ?'id-Ed448' ->
+    #'ECPrivateKey'{version = 1, parameters = {namedCurve, CurveOId}, privateKey = PrivKey};
+der_priv_key_decode(#'OneAsymmetricKey'{
+                       privateKeyAlgorithm = #'OneAsymmetricKey_privateKeyAlgorithm'{algorithm = CurveOId},
+                       privateKey = PrivKey,
+                       attributes = Attr,
+                       publicKey = PubKey}) when
+      CurveOId == ?'id-Ed25519'orelse
+      CurveOId == ?'id-Ed448' ->
+    #'ECPrivateKey'{version = 2, parameters = {namedCurve, CurveOId}, privateKey = PrivKey,
+                    attributes = Attr,
+                    publicKey = PubKey};
 der_priv_key_decode({'PrivateKeyInfo', v1,
 	{'PrivateKeyInfo_privateKeyAlgorithm', ?'rsaEncryption', _}, PrivKey, _}) ->
 	der_decode('RSAPrivateKey', PrivKey);
@@ -303,13 +365,16 @@ der_priv_key_decode({'PrivateKeyInfo', v1,
     Key = der_decode('RSAPrivateKey', PrivKey),
     Params = der_decode('RSASSA-PSS-params', Parameters),
     {Key, Params};
-der_priv_key_decode({'PrivateKeyInfo', v1,
-	{'PrivateKeyInfo_privateKeyAlgorithm', ?'id-dsa', {asn1_OPENTYPE, Parameters}}, PrivKey, _}) ->
-	{params, #'Dss-Parms'{p=P, q=Q, g=G}} = der_decode('DSAParams', Parameters),
-	X = der_decode('Prime-p', PrivKey),
-	#'DSAPrivateKey'{p=P, q=Q, g=G, x=X};
+der_priv_key_decode(#'PrivateKeyInfo'{version = v1,
+                                      privateKeyAlgorithm = #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = ?'id-dsa',
+                                                                                                  parameters =
+                                                                                                      {asn1_OPENTYPE, Parameters}},
+                                      privateKey = PrivKey}) ->
+    {params, #'Dss-Parms'{p=P, q=Q, g=G}} = der_decode('DSAParams', Parameters),
+    X = der_decode('Prime-p', PrivKey),
+    #'DSAPrivateKey'{p=P, q=Q, g=G, x=X};
 der_priv_key_decode(PKCS8Key) ->
-	PKCS8Key.
+    PKCS8Key.
 
 %%--------------------------------------------------------------------
 -spec der_encode(Asn1Type, Entity) -> Der when Asn1Type :: asn1_type(),
@@ -319,31 +384,63 @@ der_priv_key_decode(PKCS8Key) ->
 %% Description: Encodes a public key entity with asn1 DER encoding.
 %%--------------------------------------------------------------------
 der_encode('PrivateKeyInfo', #'DSAPrivateKey'{p=P, q=Q, g=G, x=X}) ->
+    Params = der_encode('Dss-Parms', #'Dss-Parms'{p=P, q=Q, g=G}),
+    Alg =  #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = ?'id-dsa',
+                                                 parameters =
+                                                     {asn1_OPENTYPE, Params}},
+    Key = der_encode('Prime-p', X),
     der_encode('PrivateKeyInfo',
-               {'PrivateKeyInfo', v1,
-                {'PrivateKeyInfo_privateKeyAlgorithm', ?'id-dsa',
-                 {asn1_OPENTYPE, der_encode('Dss-Parms', #'Dss-Parms'{p=P, q=Q, g=G})}},
-		der_encode('Prime-p', X), asn1_NOVALUE});
+               #'PrivateKeyInfo'{version = v1,
+                                 privateKeyAlgorithm = Alg,
+                                 privateKey = Key});
 der_encode('PrivateKeyInfo', #'RSAPrivateKey'{} = PrivKey) ->
+    Parms = ?DER_NULL,
+    Alg = #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = ?'rsaEncryption',
+                                                parameters = {asn1_OPENTYPE, Parms}},
+    Key = der_encode('RSAPrivateKey', PrivKey),
     der_encode('PrivateKeyInfo',
-               {'PrivateKeyInfo', v1,
-                {'PrivateKeyInfo_privateKeyAlgorithm', ?'rsaEncryption', 
-                 {asn1_OPENTYPE, ?DER_NULL}},
-                der_encode('RSAPrivateKey', PrivKey), asn1_NOVALUE});
+               #'PrivateKeyInfo'{version = v1,
+                                 privateKeyAlgorithm = Alg,
+                                 privateKey = Key});
 der_encode('PrivateKeyInfo', {#'RSAPrivateKey'{} = PrivKey, Parameters}) ->
-    der_encode('PrivateKeyInfo',
-               {'PrivateKeyInfo', v1,
-                {'PrivateKeyInfo_privateKeyAlgorithm', ?'id-RSASSA-PSS', 
-                 {asn1_OPENTYPE, der_encode('RSASSA-PSS-params', Parameters)}},
-                der_encode('RSAPrivateKey', PrivKey), asn1_NOVALUE});
+    Params = der_encode('RSASSA-PSS-params', Parameters),
+    Alg = #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = ?'id-RSASSA-PSS',
+                                                parameters =
+                                                    {asn1_OPENTYPE, Params}},
+    Key = der_encode('RSAPrivateKey', PrivKey),
+    der_encode('PrivateKeyInfo', #'PrivateKeyInfo'{version = v1,
+                                                   privateKeyAlgorithm = Alg,
+                                                   privateKey = Key});
+der_encode('PrivateKeyInfo', #'ECPrivateKey'{parameters = {namedCurve, CurveOId},
+                                             privateKey = Key}) when
+      CurveOId == ?'id-Ed25519' orelse
+      CurveOId == ?'id-Ed448' ->
+    Alg = #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = CurveOId},
+    der_encode('PrivateKeyInfo', #'PrivateKeyInfo'{version = v1,
+                                                   privateKeyAlgorithm = Alg,
+                                                   privateKey = Key});
 der_encode('PrivateKeyInfo', #'ECPrivateKey'{parameters = Parameters} = PrivKey) ->
+    Params = der_encode('EcpkParameters', Parameters),
+    Alg = #'PrivateKeyInfo_privateKeyAlgorithm'{algorithm = ?'id-ecPublicKey',
+                                                parameters = {asn1_OPENTYPE, Params}},
+    Key = der_encode('ECPrivateKey', PrivKey#'ECPrivateKey'{parameters = asn1_NOVALUE}),
     der_encode('PrivateKeyInfo',
-               {'PrivateKeyInfo', v1,
-		{'PrivateKeyInfo_privateKeyAlgorithm', ?'id-ecPublicKey',
-                 {asn1_OPENTYPE, der_encode('EcpkParameters', Parameters)}},
-                der_encode('ECPrivateKey', PrivKey#'ECPrivateKey'{parameters = asn1_NOVALUE}), 
-                asn1_NOVALUE});
-der_encode(Asn1Type, Entity) when (Asn1Type == 'PrivateKeyInfo') or
+               #'PrivateKeyInfo'{version = v1,
+                                 privateKeyAlgorithm = Alg,
+                                 privateKey = Key});
+der_encode('OneAsymmetricKey', #'ECPrivateKey'{parameters = {namedCurve, CurveOId},
+                                               privateKey = Key,
+                                               attributes = Attr,
+                                               publicKey = PubKey}) ->
+    Alg = #'OneAsymmetricKey_privateKeyAlgorithm'{algorithm = CurveOId},
+    der_encode('OneAsymmetricKey',
+               #'OneAsymmetricKey'{version = 1,
+                                   privateKeyAlgorithm = Alg,
+                                   privateKey = Key,
+                                   attributes = Attr,
+                                   publicKey = PubKey});
+der_encode(Asn1Type, Entity) when (Asn1Type == 'PrivateKeyInfo') orelse
+                                  (Asn1Type == 'OneAsymmetricKey') orelse
 				  (Asn1Type == 'EncryptedPrivateKeyInfo') ->
      try
          {ok, Encoded} = 'PKCS-FRAME':encode(Asn1Type, Entity),
@@ -587,6 +684,12 @@ generate_key({rsa, ModulusSize, PublicExponent}) ->
                              when OthersECDHkey :: #'ECPoint'{},
                                   MyECDHkey :: #'ECPrivateKey'{},
                                   SharedSecret :: binary().
+compute_key(#'ECPoint'{point = Point}, #'ECPrivateKey'{privateKey = PrivKey,
+						       parameters = {namedCurve, Curve} = Param})
+  when (Curve == ?'id-X25519') orelse
+       (Curve == ?'id-X448') ->
+    ECCurve = ec_curve_spec(Param),
+    crypto:compute_key(eddh, Point, PrivKey, ECCurve);
 compute_key(#'ECPoint'{point = Point}, #'ECPrivateKey'{privateKey = PrivKey,
 						       parameters = Param}) ->
     ECCurve = ec_curve_spec(Param),
@@ -606,7 +709,7 @@ compute_key(PubKey, PrivKey, #'DHParameter'{prime = P, base = G}) ->
                              {DigestType, SignatureType}
                                  when AlgorithmId :: oid(),
                                       %% Relevant dsa digest type is a subset of rsa_digest_type()
-                                      DigestType :: crypto:rsa_digest_type(),
+                                      DigestType :: crypto:rsa_digest_type() | none,
                                       SignatureType :: rsa | dsa | ecdsa .
 %% Description:
 %%--------------------------------------------------------------------
@@ -639,7 +742,11 @@ pkix_sign_types(?'ecdsa-with-SHA256') ->
 pkix_sign_types(?'ecdsa-with-SHA384') ->
     {sha384, ecdsa};
 pkix_sign_types(?'ecdsa-with-SHA512') ->
-    {sha512, ecdsa}.
+    {sha512, ecdsa};
+pkix_sign_types(?'id-Ed25519') ->
+    {none, eddsa};
+pkix_sign_types(?'id-Ed448') ->
+    {none, eddsa}.
 
 %%--------------------------------------------------------------------
 -spec pkix_hash_type(HashOid::oid()) -> DigestType:: md5 | crypto:sha1() | crypto:sha2().
@@ -833,10 +940,21 @@ pkix_verify(DerCert,  {#'RSAPublicKey'{} = RSAKey, #'RSASSA-PSS-params'{} = Para
     {DigestType, PlainText, Signature} = pubkey_cert:verify_data(DerCert),
     verify(PlainText, DigestType, Signature, RSAKey, rsa_opts(Params));
 
-pkix_verify(DerCert, Key = {#'ECPoint'{}, _})
-  when is_binary(DerCert) ->
-    {DigestType, PlainText, Signature} = pubkey_cert:verify_data(DerCert),
-    verify(PlainText, DigestType, Signature,  Key).
+pkix_verify(DerCert, Key = {#'ECPoint'{}, {namedCurve, Curve}}) when (Curve == ?'id-Ed25519'orelse
+                                                                      Curve == ?'id-Ed448') andalso is_binary(DerCert) ->
+    case pubkey_cert:verify_data(DerCert) of
+        {none = DigestType, PlainText, Signature} ->
+            verify(PlainText, DigestType, Signature, Key);
+        _ ->
+            false
+    end;
+pkix_verify(DerCert, Key = {#'ECPoint'{}, _}) when is_binary(DerCert) ->
+    case pubkey_cert:verify_data(DerCert) of
+        {none, _, _} ->
+            false;
+        {DigestType, PlainText, Signature} ->
+            verify(PlainText, DigestType, Signature, Key)
+    end.
 
 %%--------------------------------------------------------------------
 -spec pkix_crl_verify(CRL, Cert) -> boolean()
@@ -1366,8 +1484,13 @@ format_sign_key(Key = #'RSAPrivateKey'{}) ->
     {rsa, format_rsa_private_key(Key)};
 format_sign_key(#'DSAPrivateKey'{p = P, q = Q, g = G, x = X}) ->
     {dss, [P, Q, G, X]};
+format_sign_key(#'ECPrivateKey'{privateKey = PrivKey, parameters = {namedCurve, Curve} = Param})
+  when (Curve == ?'id-Ed25519') orelse (Curve == ?'id-Ed448')->
+    ECCurve = ec_curve_spec(Param),
+    {eddsa, [PrivKey, ECCurve]};
 format_sign_key(#'ECPrivateKey'{privateKey = PrivKey, parameters = Param}) ->
-    {ecdsa, [PrivKey, ec_curve_spec(Param)]};
+    ECCurve = ec_curve_spec(Param),
+    {ecdsa, [PrivKey, ECCurve]};
 format_sign_key({ed_pri, Curve, _Pub, Priv}) ->
     {eddsa, [Priv,Curve]};
 format_sign_key(_) ->
@@ -1375,8 +1498,13 @@ format_sign_key(_) ->
 
 format_verify_key(#'RSAPublicKey'{modulus = Mod, publicExponent = Exp}) ->
     {rsa, [Exp, Mod]};
+format_verify_key({#'ECPoint'{point = Point}, {namedCurve, Curve} = Param}) when (Curve == ?'id-Ed25519') orelse
+                                                                                 (Curve == ?'id-Ed448') ->
+    ECCurve = ec_curve_spec(Param),
+    {eddsa, [Point, ECCurve]};
 format_verify_key({#'ECPoint'{point = Point}, Param}) ->
-    {ecdsa, [Point, ec_curve_spec(Param)]};
+    ECCurve = ec_curve_spec(Param),
+    {ecdsa, [Point, ECCurve]};
 format_verify_key({Key,  #'Dss-Parms'{p = P, q = Q, g = G}}) ->
     {dss, [P, Q, G, Key]};
 format_verify_key({ed_pub, Curve, Key}) ->
@@ -1628,7 +1756,8 @@ format_rsa_private_key(#'RSAPrivateKey'{modulus = N, publicExponent = E,
 -spec ec_generate_key(ecpk_parameters_api()) -> #'ECPrivateKey'{}.
 ec_generate_key(Params) ->
     Curve = ec_curve_spec(Params),
-    Term = crypto:generate_key(ecdh, Curve),
+    CurveType = ec_curve_type(Curve),
+    Term = crypto:generate_key(CurveType, Curve),
     NormParams = ec_normalize_params(Params),
     ec_key(Term, NormParams).
 
@@ -1646,16 +1775,31 @@ ec_curve_spec( #'ECParameters'{fieldID = #'FieldID'{fieldType = Type,
     Curve = {PCurve#'Curve'.a, PCurve#'Curve'.b, none},
     {Field, Curve, Base, Order, CoFactor};
 ec_curve_spec({ecParameters, ECParams}) ->
-	ec_curve_spec(ECParams);
+    ec_curve_spec(ECParams);
 ec_curve_spec({namedCurve, OID}) when is_tuple(OID), is_integer(element(1,OID)) ->
     ec_curve_spec({namedCurve,  pubkey_cert_records:namedCurves(OID)});
 ec_curve_spec({namedCurve, x25519 = Name}) ->
     Name;
 ec_curve_spec({namedCurve, x448 = Name}) ->
     Name;
+ec_curve_spec({namedCurve, ed25519 = Name}) ->
+    Name;
+ec_curve_spec({namedCurve, ed448 = Name}) ->
+    Name;
 ec_curve_spec({namedCurve, Name}) when is_atom(Name) ->
     crypto:ec_curve(Name).
 
+ec_curve_type(ed25519) ->
+    eddsa;
+ec_curve_type(ed448) ->
+    eddsa;
+ec_curve_type(x25519) ->
+    eddh;
+ec_curve_type(x448) ->
+    eddh;
+ec_curve_type(_) ->
+    ecdh.
+
 format_field(characteristic_two_field = Type, Params0) ->
     #'Characteristic-two'{
        m = M,
@@ -1901,3 +2045,6 @@ ocsp_status(Cert, IssuerCert, Responses) ->
 ocsp_responses(OCSPResponseDer, ResponderCerts, Nonce) ->
     pubkey_ocsp:verify_ocsp_response(OCSPResponseDer, 
                                      ResponderCerts, Nonce).
+
+subject_public_key_info(Alg, PubKey) ->
+    #'OTPSubjectPublicKeyInfo'{algorithm = Alg, subjectPublicKey = PubKey}.
